import unittest

from bddrest.cli import main
from bddrest.tests.helpers import standard_files_mockup


def test_document_cli():
    with standard_files_mockup(yamlstory, argv=['bddrest', 'document']) as (stdout, stderr):
        main()

    assert expected_markdown == stdout.getvalue().decode()


def test_help():
    with standard_files_mockup(yamlstory, argv=['bddrest']) \
            as (stdout, stderr):
        main()

    assert expected_help == stdout.getvalue().decode()


yamlstory = '''
base_call:
  as_: visitor
  description: As a member I have to POST a book to the library.
  form:
    name: BDD Book
  query:
    a: b
  response:
    headers:
    - 'Content-Type: application/json;charset=utf-8'
    json:
      foo: bar
    status: 200 OK
  title: Posting a book
  url: /books/:id
  url_parameters:
    id: '1'
  verb: GET
calls:
- response:
    headers:
    - 'Content-Type: text/plain;charset=utf-8'
    status: 404 Not Found
  title: Trying invalid book id
  url_parameters:
    id: None
'''

expected_markdown = '''\
## Posting a book

### GET /books/:id

As a member I have to POST a book to the library.

### Url Parameters

Name | Example
--- | ---
id | 1

### Query Strings

Name | Example
--- | ---
a | b

### Form

Name | Example
--- | ---
name | BDD Book

### Response: 200 OK

#### Headers

* Content-Type: application/json;charset=utf-8

#### Body

```json
{"foo": "bar"}
```

## WHEN: Trying invalid book id

### Url Parameters

Name | Example
--- | ---
id | None

### Response: 404 Not Found

#### Headers

* Content-Type: text/plain;charset=utf-8

'''


expected_help = '''\
usage: bddrest [-h] {document} ...

bddrest command line interface.

optional arguments:
  -h, --help  show this help message and exit

Sub Commands:
  {document}
    document  Generates REST API Documentation from standard input to standard
              output.
'''



from ..context import Context
from ..specification import FirstCall, AlteredCall, Call
from .story import Story


class Given(Story, Context):
    """
    :param application: A WSGI Application to examine
    :param autodump: A string which indicates the filename to dump the story, or
                     a `callable(story) -> filename` to determine the filename.
                     A file-like object is also accepted.
                     Default is `None`, means autodump is disabled by default.
    :param autodoc: A string which indicates the name of documentation file, or
                     a `callable(story) -> filename` to determine the filename.
                     A file-like object is also accepted.
                     Default is `None`, meana autodoc is disabled by default.
                     Currently only markdown is supprted.
    """

    def __init__(self, application, *args, autodump=None, autodoc=None, **kwargs):
        self.application = application
        self.autodump = autodump
        self.autodoc = autodoc
        base_call = FirstCall(*args, **kwargs)
        base_call.conclude(application)
        super().__init__(base_call)

    @property
    def current_call(self) -> Call:
        if self.calls:
            return self.calls[-1]
        else:
            return self.base_call

    def when(self, *args, **kwargs):
        new_call = AlteredCall(self.base_call, *args, **kwargs)
        new_call.conclude(self.application)
        self.calls.append(new_call)
        return new_call

    def __enter__(self):
        return super().__enter__()

    def __exit__(self, exc_type, exc_value, traceback):
        super().__exit__(exc_type, exc_value, traceback)
        if self.autodump:
            if hasattr(self.autodump, 'write'):
                self.dump(self.autodump)
            else:
                filename = self.autodump(self) if callable(self.autodump) else self.autodump
                with open(filename, mode='w', encoding='utf-8') as f:
                    self.dump(f)

        if self.autodoc:
            if hasattr(self.autodoc, 'write'):
                self.dump(self.autodoc)
            else:
                filename = self.autodoc(self) if callable(self.autodoc) else self.autodoc
                with open(filename, mode='w', encoding='utf-8') as f:
                    self.document(f)

    @property
    def response(self):
        if self.current_call is None:
            return None
        return self.current_call.response





import contextlib
import io
import sys


@contextlib.contextmanager
def standard_files_mockup(stdin, argv=None):

    class StandardFile:
        encoding = 'UTF-8'
        def __init__(self, content=None):
            if content is None:
                self.buffer = io.BytesIO()
            else:
                self.buffer = io.BytesIO(content.encode() if isinstance(content, str) else content)

        def write(self, d):
            self.buffer.write(d.encode())

        def read(self, i=None):
            return self.buffer.read(i).decode()

        def seek(self, n):
            self.buffer.seek(0)

        def getvalue(self):
            return self.buffer.getvalue()

    stdinfile = StandardFile(stdin)
    stdoutfile = StandardFile()
    stderrfile = StandardFile()

    stdinbackup = sys.stdin
    stdoutbackup = sys.stdout
    stderrbackup = sys.stderr
    argvbackup = sys.argv

    sys.stdin = stdinfile
    sys.stdout = stdoutfile
    sys.stderr = stderrfile
    if argv:
        sys.argv = argv

    yield stdoutfile, stderrfile

    sys.stdin = stdinbackup
    sys.stdout = stdoutbackup
    sys.stderr = stderrbackup
    if argv:
        sys.argv = argvbackup

    stdoutfile.seek(0)
    stderrfile.seek(0)




import sys
import argparse

from .formatters import *
from .documenter import Documenter
from ..cli import Launcher


class DocumentaryLauncher(Launcher):
    formatters = {
        'markdown': MarkdownFormatter,
        # 'html': HtmlFormatter,
    }

    @classmethod
    def create_parser(cls, subparsers):
        parser = subparsers.add_parser(
            'document',
            help='Generates REST API Documentation from standard input to standard output.'
        )
        parser.add_argument(
            '-f',
            '--format',
            default='markdown',
            help='The output format. One of markdown, html. Default is markdown.'
        )
        return parser

    def convert_file(self, source, destination):
        from ..authoring import Story
        story = Story.load(source)
        story.document(destination, formatter_factory=self.formatters[self.args.format])

    def launch(self):
        self.convert_file(sys.stdin, sys.stdout)




import cgi
import json
import pytest

from bddrest import Given, Append, Remove, Update, when, response


def wsgi_application(environ, start_response):
    form = cgi.FieldStorage(
        fp=environ['wsgi.input'],
        environ=environ,
        strict_parsing=False,
        keep_blank_values=True
    )

    start_response('200 OK', [
        ('Content-Type', 'application/json;charset=utf-8'),
    ])
    yield json.dumps({k: form[k].value for k in form.keys()}).encode()


def test_append_form_field():
    call = dict(
        title='test add form field',
        url='/apiv1/devices/name: SM-12345678/id: 1',
        verb='POST',
        form=dict(
            activationCode='746727',
            phone='+9897654321'
        ),
    )

    with Given(wsgi_application, **call):
        assert response.status == '200 OK'

        modified_call = when(
            'Adding another field',
            form=Append(email='user@example.com')
        )
        assert response.json == dict(
            activationCode='746727',
            phone='+9897654321',
            email='user@example.com'
        )


def test_remove_from_fields():
    call = dict(
            title='test remove form fields',
            url='/apiv1/devices/name: SM-12345678/id: 1',
            verb='POST',
            form=dict(
                activationCode='746727',
                phone='+9897654321',
                email='user@example.com'
            )
    )

    with Given(wsgi_application, **call):
        assert response.status =='200 OK'

        when('Removing fields', form=Remove('email', 'phone'))
        assert response.json == dict(activationCode='746727')

        with pytest.raises(ValueError):
            Remove('a').apply(['b', 'c'])

        with pytest.raises(ValueError):
            Remove('a').apply({'b': 'c'})


def test_update_from_fields():
    call = dict(
            title='test remove form fields',
            url='/apiv1/devices/name: SM-12345678/id: 1',
            verb='POST',
            form=dict(
                activationCode='746727',
                email='user@example.com'
            )
    )

    with Given(wsgi_application, **call):
        assert response.status =='200 OK'

        when(
            'Updating fields',
            form=Update(email='test@example.com', phone='+98123456789')
        )
        assert response.json == dict(
            activationCode='746727',
            phone='+98123456789',
            email='test@example.com'
        )




import unittest
import re

from bddrest import HeaderSet


def test_constructor():
    expected_headers = [('A', 'B'), ('C', 'D')]
    headers = HeaderSet(('A: B', 'C: D'))
    assert expected_headers == headers

    headers = HeaderSet({'A': 'B', 'C': 'D'})
    assert expected_headers == headers

def test_append():
    expected_headers = [('A', 'B'), ('C', 'D')]
    headers = HeaderSet()
    headers.append(('A', 'B'))
    headers.append(('C', 'D'))
    assert expected_headers == headers

    headers = HeaderSet()
    headers.append('A: B')
    headers.append('C: D')
    assert expected_headers == headers

    headers = HeaderSet()
    headers.append('A', 'B')
    headers.append('C', 'D')
    assert expected_headers == headers

def test_insert():
    expected_headers = [('A', 'B'), ('C', 'D')]
    headers = HeaderSet()
    headers.insert(0, ('C', 'D'))
    headers.insert(0, ('A', 'B'))
    assert expected_headers == headers

    headers = HeaderSet()
    headers.insert(0, 'C: D')
    headers.insert(0, 'A: B')
    assert expected_headers == headers

    headers = HeaderSet()
    headers.insert(0, 'C', 'D')
    headers.insert(0, 'A', 'B')
    assert expected_headers == headers

def test_setitem_getitem_delitem():
    expected_headers = [('A', 'B'), ('C', 'D')]
    # setitem
    headers = HeaderSet(['A: F'])
    headers[0] = 'A: B'
    headers['C'] = 'D'
    assert expected_headers == headers

    # getitem
    assert expected_headers[0] == headers[0]
    assert 'B' == headers['A']
    assert 'B' == headers['a']

    # delitem
    del headers[0]
    del headers['C']
    assert [] == headers

def test_in_operator():
    headers = HeaderSet(['A: F'])
    assert 'A' in headers
    assert 'a' in headers
    assert 'A: F' in headers

def test_extend():
    expected_headers = [('A', 'B'), ('C', 'D')]

    headers = HeaderSet(['A: B'])
    headers.extend(['C: D'])
    assert expected_headers == headers

def test_regex_match():
    headers = HeaderSet(['Content-Type: application/json;utf-8'])
    assert re.compile('^content-type: .*', re.I) in headers

def test_remove():
    headers = HeaderSet([('A', 'B'), ('C', 'D')])
    headers.remove('A')
    headers.remove('C: D')
    assert [] == headers




import yaml

from ..documentary import Documenter, MarkdownFormatter
from ..specification import FirstCall, AlteredCall, Call


class Story:
    _yaml_options = dict(default_style=False, default_flow_style=False)

    def __init__(self, base_call, calls=None):
        self.base_call = base_call
        self.calls = calls or []

    def to_dict(self):
        return dict(
            base_call=self.base_call.to_dict(),
            calls=[c.to_dict() for c in self.calls]
        )

    @classmethod
    def from_dict(cls, data):
        base_call = FirstCall(**data['base_call'])
        return cls(
            base_call,
            calls=[
                AlteredCall(base_call, **d)
                for d in data['calls']
            ] if data.get('calls') else None
        )

    def dump(self, file):
        data = self.to_dict()
        yaml.dump(data, file, **self._yaml_options)

    def dumps(self):
        data = self.to_dict()
        return yaml.dump(data, **self._yaml_options)

    def verify(self, application):
        self.base_call.verify(application)
        for c in self.calls:
            c.verify(application)

    @classmethod
    def load(cls, file):
        data = yaml.load(file)
        return cls.from_dict(data)

    @classmethod
    def loads(cls, string):
        data = yaml.load(string)
        return cls.from_dict(data)

    def validate(self):
        self.base_call.validate()
        for call in self.calls:
            call.validate()

    def document(self, outfile, formatter_factory=MarkdownFormatter):
        documenter = Documenter(formatter_factory)
        documenter.document(self, outfile)

    @property
    def title(self):
        return self.base_call.title




import re
import json as jsonlib

from .headerset import HeaderSet


CONTENT_TYPE_PATTERN = re.compile('(\w+/\w+)(?:;\s?charset=(.+))?')


class HTTPStatus:

    def __init__(self, code):
        self.code = int(code.split(' ', 1)[0])
        self.text = code

    def raise_value_error(self):
        raise ValueError(
            'Cannot compare with string, Use integer instead for all '
            'comparison types except equality'
        )

    def __eq__(self, other):
        if isinstance(other, int):
            return self.code == other

        if isinstance(other, self.__class__):
            other = other.text

        return self.text.casefold() == other.casefold()

    def __gt__(self, other):
        if isinstance(other, int):
            return self.code > other
        self.raise_value_error()

    def __ge__(self, other):
        if isinstance(other, int):
            return self.code >= other
        self.raise_value_error()

    def __lt__(self, other):
        if isinstance(other, int):
            return self.code < other
        self.raise_value_error()

    def __le__(self, other):
        if isinstance(other, int):
            return self.code <= other
        self.raise_value_error()

    def __str__(self):
        return self.text

    def __repr__(self):
        return f'\'{str(self)}\''


class Response:
    content_type = None
    encoding = None
    body = None

    def __init__(self, status, headers, body=None, json=None):
        self.status = HTTPStatus(status)
        self.headers = HeaderSet(headers) if headers is not None else None
        if json:
            self.body = jsonlib.dumps(json).encode()
            # FIXME: enable it after HeaderSet is implemented.
            # self.headers.append('Content-Type: application/json;charset=utf-8')
        elif body:
            self.body = body.encode() if not isinstance(body, bytes) else body

        if headers:
            for k, v in self.headers:
                if k == 'Content-Type':
                    match = CONTENT_TYPE_PATTERN.match(v)
                    if match:
                        self.content_type, self.encoding = match.groups()
                    break

    @property
    def text(self):
        return self.body.decode()

    @property
    def json(self):
        return jsonlib.loads(self.body)

    def to_dict(self):
        result = dict(
            status=str(self.status)
        )
        if self.headers:
            result['headers'] = self.headers.simple

        if self.body:
            if self.content_type == 'application/json':
                result['json'] = self.json
            else:
                result['body'] = self.body.decode()
        return result

    def __eq__(self, other: 'Response'):
        if self.status != other.status or self.headers != other.headers:
            return False

        if self.content_type == 'application/json':
            return self.json == other.json

        return self.body == other.body




from urllib.parse import parse_qs


def normalize_query_string(query):
    if not query:
        return None
    return {
        k: v[0] if len(v) == 1 else v for k, v in parse_qs(query).items()
    } if isinstance(query, str) else query




import re
from typing.re import Pattern

from collections import Iterable


class HeaderSet(list):
    def __init__(self, headers=None):
        if headers:
            if isinstance(headers, dict):
                headers = list(headers.items())

            super().__init__(self._normalize_item(i) for i in headers)
        else:
            super().__init__()

    @staticmethod
    def _normalize_item(h):
        k, v = h.split(':', 1) if isinstance(h, str) else h
        return k, v.strip()

    @property
    def simple(self):
        return [f'{k}: {v}' for k, v in self]

    def _get_item_by_key(self, key):
        testkey = key.casefold()
        if ':' in key:
            for i, h in enumerate(self.simple):
                if testkey == h.casefold():
                    k, v = self[i]
                    return i, k, v
        else:
            for i, (k, v) in enumerate(self):
                if k.casefold() == testkey:
                    return i, k, v

        raise KeyError(key)

    def append(self, k, v=None):
        if v:
            return super().append((k, v))
        else:
            return super().append(self._normalize_item(k))

    def insert(self, i, k, v=None):
        if v:
            return super().insert(i, (k, v))
        else:
            return super().insert(i, self._normalize_item(k))

    def __setitem__(self, key, value):
        if isinstance(key, str):
            self.append(key, value)
        else:
            super().__setitem__(key, self._normalize_item(value))

    def __getitem__(self, key):
        if isinstance(key, str):
            return self._get_item_by_key(key)[2]
        else:
            return super().__getitem__(key)

    def __delitem__(self, key):
        if isinstance(key, str):
            super().__delitem__(self._get_item_by_key(key)[0])
        else:
            super().__delitem__(key)

    def __contains__(self, key):
        if isinstance(key, str):
            try:
                self._get_item_by_key(key)
                return True
            except KeyError:
                return False
        elif isinstance(key, Pattern):
            for i in self.simple:
                if key.match(i):
                    return True
            return False
        else:
            return super().__contains__(key)

    def extend(self, other):
        super().extend(self._normalize_item(i) for i in other)

    def remove(self, key):
        if isinstance(key, str):
            key = self._get_item_by_key(key)[1:]
        super().remove(key)




import cgi
import functools
import json
import unittest

import pytest
from bddrest import CallVerifyError, FirstCall, AlteredCall


def wsgi_application(environ, start_response):
    form = cgi.FieldStorage(
        fp=environ['wsgi.input'],
        environ=environ,
        strict_parsing=False,
        keep_blank_values=True
    )

    start_response(
        '200 OK',
        [('Content-Type', 'application/json;charset=utf-8')]
    )
    result = dict(
        query=environ['QUERY_STRING'],
        url=environ['PATH_INFO']
    )
    if form and isinstance(form, dict):
        result.update(form)
    yield json.dumps(result).encode()


def test_call_constructor():
    call = FirstCall('Testing Call contractor', url='/id: 1')
    assert call.url == '/:id'
    assert call.url_parameters == dict(id='1')

    call = FirstCall(
        'Testing Call contractor',
        url='/id: 1/:name',
        url_parameters=dict(name='foo', id=2)
    )
    call.validate()
    assert call.url == '/:id/:name'
    assert call.url_parameters == dict(id='2', name='foo')
    call.conclude(wsgi_application)
    assert '/2/foo' == call.response.json['url']


def test_call_invoke():
    call = FirstCall('Testing Call contractor', url='/id: 1')
    call.conclude(wsgi_application)
    assert call.response is not None


def test_call_response():
    call = FirstCall('Testing Call contractor', url='/id: 1', query='a=1')
    call.conclude(wsgi_application)
    assert call.response is not None
    assert call.response.body is not None
    assert call.response.status == '200 OK'
    assert call.response.status == 200
    assert call.response.encoding == 'utf-8'
    assert call.response.content_type == 'application/json'
    assert call.response.text is not None
    assert call.response.json == {'query': 'a=1', 'url': '/1'}
    assert call.response.headers == [
        ('Content-Type', 'application/json;charset=utf-8')
    ]


def test_call_to_dict():
    call = FirstCall('Testing Call to_dict', url='/id: 1', query='a=1')
    call.conclude(wsgi_application)
    call_dict = call.to_dict()
    assert call_dict == dict(
        title='Testing Call to_dict',
        query=dict(a='1'),
        url='/:id',
        url_parameters={'id': '1'},
        verb='GET',
        response=dict(
            json={'query': 'a=1', 'url': '/1'},
            headers=['Content-Type: application/json;charset=utf-8'],
            status='200 OK',
        )
    )


def test_altered_call():
    call = FirstCall(
        'Testing AlteredCall contractor',
        url='/id: 1',
        query=dict(a=1)
    )

    altered_call = AlteredCall(
        call,
        'Altering a call',
        query=dict(b=2)
    )
    altered_call.conclude(wsgi_application)
    assert altered_call.to_dict() == dict(
        title='Altering a call',
        query=dict(b=2),
        response=dict(
            status='200 OK',
            headers=['Content-Type: application/json;charset=utf-8'],
            json={'query': 'b=2', 'url': '/1'}
        )
    )


def test_alteredcall_setters_deleters():
    basecall = FirstCall(
        'Base call for testing When class',
        url='/apiv1/devices/id: 1',
    )

    when = AlteredCall(
        basecall,
        title='Testing the When class',
        url='/apiv1/books/isbn: abc/pages/page: 3?highlight=false',
        verb='POST',
        form=dict(a='b'),
        headers=['A: B'],
        content_type='text/plain',
        as_='Admin',
        extra_environ=dict(A='B')
    )
    assert '/apiv1/books/:isbn/pages/:page' == when.url
    assert dict(isbn='abc', page='3') == when.url_parameters
    assert dict(highlight='false') == when.query
    assert dict(a='b') == when.form
    assert 'POST' == when.verb
    assert 'A' in when.headers
    assert 'text/plain' == when.content_type
    assert 'Admin' == when.as_
    del when.url_parameters
    del when.verb
    del when.headers
    del when.query
    del when.content_type
    del when.as_
    del when.extra_environ
    del when.form

    assert dict(id='1') == when.url_parameters
    assert 'GET' == when.verb
    assert when.headers is None
    assert when.query is None
    assert when.form is None
    assert when.content_type is None
    assert when.as_ is None
    assert when.extra_environ is None


def test_call_verify():
    call = FirstCall(
        'Testing FirstCall contractor',
        url='/id: 1',
        query=dict(a=1)
    )

    call.conclude(wsgi_application)
    call.verify(wsgi_application)

    altered_call = AlteredCall(
        call,
        'Altering a call',
        query=dict(b=2)
    )
    altered_call.conclude(wsgi_application)
    altered_call.verify(wsgi_application)

    altered_call.response.body = '{"a": 1}'
    with pytest.raises(CallVerifyError):
        altered_call.verify(wsgi_application)

    altered_call.response.status = '400 Bad Request'
    with pytest.raises(CallVerifyError):
        altered_call.verify(wsgi_application)


def test_querystring_parser():
    call = FirstCall('Testing querystring parsing', url='/id: 1?a=1')
    assert '/:id' == call.url
    assert dict(a='1') == call.query


def test_form_parser():
    pyload = dict(a=1, b=2)
    call = FirstCall('Testing form parsing', form=pyload)
    assert call.form == pyload




import re
import sys
import json
from abc import ABCMeta, abstractmethod
from urllib.parse import urlparse, urlencode

from webtest import TestApp

from ..helpers import normalize_query_string
from ..exceptions import CallVerifyError, InvalidUrlParametersError
from .response import Response


URL_PARAMETER_VALUE_PATTERN = '[\w\d_-]+'
URL_PARAMETER_PATTERN = \
    re.compile(f'/(?P<key>\w+):\s?(?P<value>{URL_PARAMETER_VALUE_PATTERN})')


class Call(metaclass=ABCMeta):

    def __init__(self, title, description=None, response=None):
        self.title = title
        self.description = description
        if response is not None and not isinstance(response, Response):
            response = Response(**response)
        self.response = response

    def to_dict(self):
        result = dict(
            title=self.title,
            url=self.url,
            verb=self.verb,
        )
        if self.url_parameters is not None:
            result['url_parameters'] = self.url_parameters

        if self.form is not None:
            result['form'] = self.form

        if self.headers is not None:
            result['headers'] = [': '.join(h) for h in self.headers]

        if self.as_ is not None:
            result['as_'] = self.as_

        if self.query is not None:
            result['query'] = self.query

        if self.description is not None:
            result['description'] = self.description

        if self.response is not None:
            result['response'] = self.response.to_dict()

        return result

    def validate_url_parameters(self):
        required_parameters = set(i[1:] for i in re.findall(':\w+', self.url))
        if not required_parameters and self.url_parameters is None:
            return

        given_parameters = set(self.url_parameters or [])

        if given_parameters != required_parameters:
            raise InvalidUrlParametersError(
                required_parameters,
                given_parameters
            )

        for k, v in self.url_parameters.items():
            if not isinstance(v, str):
                self.url_parameters[k] = str(v)

    def validate(self):
        self.validate_url_parameters()

    @staticmethod
    def extract_url_parameters(url):
        url_parameters = {}
        query = None
        parsedurl = urlparse(url)

        # Parsing the querystrings if available
        if parsedurl.query:
            query = normalize_query_string(parsedurl.query)

        url = parsedurl.path
        if URL_PARAMETER_PATTERN.search(url):
            for k, v in URL_PARAMETER_PATTERN.findall(url):
                url_parameters[k] = v
                url = re.sub(f'{k}:\s?{URL_PARAMETER_VALUE_PATTERN}', f':{k}', url)

        return url, url_parameters if url_parameters else None, query

    def add_header_if_not_exists(self, headers, key, value):
        for k, v in headers:
            if k.lower() == key.lower():
                headers.remove((k, v))
        headers.append((key, value))

    def invoke(self, application) -> Response:
        url = self.url
        if self.url_parameters:
            for k, v in self.url_parameters.items():
                url = url.replace(f':{k}', str(v))

        url = f'{url}?{urlencode(self.query)}' if self.query else url

        headers = self.headers or []
        if self.content_type:
            self.add_header_if_not_exists(
                headers,
                'Content-Type',
                self.content_type
            )

        if self.authorization:
            self.add_header_if_not_exists(
                headers,
                'authorization',
                self.authorization
            )

        request_params = dict(
            expect_errors=True,
            extra_environ=self.extra_environ,
            headers=headers,
            # Commented for future usages by pylover
            # upload_files=upload_files,
        )
        if self.form:
            request_params['params'] = json.dumps(self.form) \
                if self.content_type and \
                self.content_type.startswith('application/json') \
                else self.form

        # noinspection PyProtectedMember
        web_test_response = TestApp(application)._gen_request(
            self.verb,
            url,
            **request_params
        )

        response = Response(
            web_test_response.status,
            [(k, v) for k, v in web_test_response.headers.items()],
            body=web_test_response.body
        )

        if 500 <= response.status < 600:
            print(response.json['stackTrace'], file=sys.stderr)

        return response


    def verify(self, application):
        response = self.invoke(application)
        if self.response != response:
            raise CallVerifyError()

    def conclude(self, application):
        if self.response is None:
            self.validate()
            self.response = self.invoke(application)

    @property
    @abstractmethod
    def verb(self) -> str:  # pragma: no cover
        pass

    @verb.setter
    @abstractmethod
    def verb(self, value):  # pragma: no cover
        pass

    @property
    @abstractmethod
    def url(self) -> str:  # pragma: no cover
        pass

    @url.setter
    @abstractmethod
    def url(self, value):  # pragma: no cover
        pass

    @property
    @abstractmethod
    def url_parameters(self) -> dict:  # pragma: no cover
        pass

    @url_parameters.setter
    @abstractmethod
    def url_parameters(self, value):  # pragma: no cover
        pass

    @property
    @abstractmethod
    def headers(self):  # pragma: no cover
        pass

    @headers.setter
    @abstractmethod
    def headers(self, value):  # pragma: no cover
        pass

    @property
    @abstractmethod
    def form(self) -> dict:  # pragma: no cover
        pass

    @form.setter
    @abstractmethod
    def form(self, value):  # pragma: no cover
        pass

    @property
    @abstractmethod
    def query(self) -> dict:  # pragma: no cover
        pass

    @query.setter
    @abstractmethod
    def query(self, value):  # pragma: no cover
        pass

    @property
    @abstractmethod
    def content_type(self) -> str:  # pragma: no cover
        pass

    @content_type.setter
    @abstractmethod
    def content_type(self, value):  # pragma: no cover
        pass

    @property
    @abstractmethod
    def authorization(self) -> str:  # pragma: no cover
        pass

    @content_type.setter
    @abstractmethod
    def authorization(self, value):  # pragma: no cover
        pass

    @property
    @abstractmethod
    def as_(self) -> str:  # pragma: no cover
        pass

    @as_.setter
    @abstractmethod
    def as_(self, value):  # pragma: no cover
        pass

    @property
    @abstractmethod
    def extra_environ(self) -> dict:  # pragma: no cover
        pass

    @extra_environ.setter
    @abstractmethod
    def extra_environ(self, value):  # pragma: no cover
        pass




import io

from bddrest import Story


def test_markdown():
    story = Story.loads(provided_story)
    outfile = io.StringIO()
    story.document(outfile)
    outputstring = outfile.getvalue()
    assert expected_markdown == outputstring


provided_story = '''
    base_call:
      as_: visitor
      title: Quickstart!
      description: Awesome API!
      url: /books/:id
      url_parameters:
        id: '1'
      query:
        a: 1
        b: '2'
      verb: PUT
      form:
        f1: abc
        f2: 123
      headers:
        - 'Content-Type: application/json;charset=utf-8'
      response:
        headers:
        - 'Content-Type: application/json;charset=utf-8'
        json:
          foo: bar
        status: 200 OK
    calls:
    - response:
        headers:
        - 'Content-Type: text/plain;charset=utf-8'
        status: 404 Not Found
      title: Trying invalid book id
      description: trying an invalid book id.
      url_parameters:
        id: None
      query:
        a: 2
        b: 4
      headers:
      - 'A: B'
      form:
        f1: cba
'''

expected_markdown = '''\
## Quickstart!

### PUT /books/:id

Awesome API!

### Url Parameters

Name | Example
--- | ---
id | 1

### Query Strings

Name | Example
--- | ---
a | 1
b | 2

### Form

Name | Example
--- | ---
f1 | abc
f2 | 123

### Request Headers

* Content-Type: application/json;charset=utf-8

### Response: 200 OK

#### Headers

* Content-Type: application/json;charset=utf-8

#### Body

```json
{"foo": "bar"}
```

## WHEN: Trying invalid book id

trying an invalid book id.

### Url Parameters

Name | Example
--- | ---
id | None

### Query Strings

Name | Example
--- | ---
a | 2
b | 4

### Form

Name | Example
--- | ---
f1 | cba

### Request Headers

* A: B

### Response: 404 Not Found

#### Headers

* Content-Type: text/plain;charset=utf-8

'''





from ..proxy import ObjectProxy
from ..specification import FirstCall, AlteredCall, Call

from .story import Story
from .given import Given
from .manipulation import Manipulator, Append, Remove, Update, when


story = ObjectProxy(Given.get_current)
response = ObjectProxy(lambda: story.response)
status = ObjectProxy(lambda: response.status)





from .specification import Call, FirstCall, AlteredCall, HeaderSet, HTTPStatus
from .authoring import Given, when, story, response, Story, Append, Update,\
    Remove, status
from .exceptions import InvalidUrlParametersError, CallVerifyError


__version__ = '1.11.0a6'

